extern crate proc_macro;

use core::panic;

use proc_macro::TokenStream;
use syn::ItemFn;
use quote::quote;

/// Sets up the given `async fn` to be the entry point.
/// 
/// ```no_run
/// # // i see no way to `use messing_with_async::block_on`, so we cant run this test
/// #[async_main]
/// async fn async_main() {
///     println!("Hello, world!");
/// }
/// ```
#[proc_macro_attribute]
pub fn async_main(_: TokenStream, item: TokenStream) -> TokenStream {
    let func = syn::parse::<ItemFn>(item).unwrap_or_else(
        |_| panic!("`async_main` can only be applied to functions!")
    );
    
    if func.sig.asyncness.is_none() {
        panic!("`async_main` can only be applied to an `async fn`!");
    }

    if !func.sig.inputs.is_empty() {
        panic!("`async_main` cannot be applied to functions with arguments!");
    }

    if !func.sig.generics.params.is_empty() {
        panic!("`async_main` cannot be applied to generic functions!");
    }

    if func.sig.unsafety.is_some() {
        panic!("`async_main` cannot be applied to unsafe functions!");
    }

    let ident = func.sig.ident.clone();

    quote! {
        fn main() {
            use ::messing_with_async::block_on;

            #func

            let _ = block_on(#ident());
        }
    }.into()
}